/* Pxys - An IRC ProxyScan, conf.c
 * 
 * Copyright (C) 2001,2002 Stephane Thiell <mbuna@undernet.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * 
 * Nothing here really needs to be fast, but every errors should be detected.
 *
 * 
 * $Id: conf.c,v 1.27 2002/12/04 15:59:06 mbuna Exp $
 */


#include "config.h"

#include <stdio.h>

#include <string.h>
#ifdef HAVE_CTYPE_H
#include <ctype.h>
#endif
#include <stdlib.h>

#include <unistd.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>

#include "conf.h"
#include "debug.h"
#include "misc.h"
#include "ms_globals.h"
#include "m_structs.h"
#include "strnlwrcpy.h"

/* xmalloc at the bottom of include files */
#include "xmalloc.h"


#ifndef INADDR_NONE
#define INADDR_NONE   0xffffffff
#endif


/* Usefull macro */
#define	checkZone(z)					\
	do {								\
		if (z != confZone) {			\
			fclose(fp);					\
			syntaxErr(line_counter);	\
			return false;				\
		}								\
	} while (false)

#define checkDouble(o1, o2)				\
	do {								\
		if(o1 & o2) {					\
			fclose(fp);					\
			printf("Configuration file: Illegal double option (line %d).\n", line_counter);	\
			return false;				\
		}								\
	} while(false)						\

#define CONFZONE_WORD_ME			"me"
#define CONFZONE_WORD_UPLINK		"uplink"
#define CONFZONE_WORD_CONSOLE		"console"
#define CONFZONE_WORD_CLIENT		"client"
#define CONFZONE_WORD_PERFORM		"perform"
#define CONFZONE_WORD_SHARE			"share"
#define CONFZONE_WORD_NOSCAN		"noscan"
#define CONFZONE_WORD_LOG			"log"
#define CONFZONE_WORD_TARGET		"target"
#define CONFZONE_WORD_GLINE			"gline"

const unsigned int confZone_nil = 0x0000;
const unsigned int confZone_me = 0x0001;
const unsigned int confZone_uplink = 0x0002;
const unsigned int confZone_console = 0x0004;
const unsigned int confZone_client = 0x0008;
const unsigned int confZone_perform = 0x0010;
const unsigned int confZone_share = 0x0020;
const unsigned int confZone_noscan = 0x0040;
const unsigned int confZone_log = 0x0080;
const unsigned int confZone_target = 0x0100;
const unsigned int confZone_gline = 0x0200;

struct cfg_Var_inZone
{
   char *vStr;
   unsigned int vID;
};

/* "me" zone variables */
static struct cfg_Var_inZone cZ_me_vlist[] = {
   {"name", 0x01},
   {"numeric", 0x02},
   {"maxclients", 0x04},
   {"serverinfo", 0x08},
   {"hostip", 0x10},
   {"bindip", 0x20},
   {"listener", 0x40},
   {"id", 0x80},
   {NULL, 0}
};

/* "uplink" zone variables */
static struct cfg_Var_inZone cZ_uplink_vlist[] = {
   {"host", 0x01},
   {"bindip", 0x02},
   {"port", 0x04},
   {"password", 0x08},
   {"enabled", 0x10},
   {"priority", 0x20},
   {NULL, 0}
};

/* "console" zone variables */
static struct cfg_Var_inZone cZ_console_vlist[] = {
   {"channel", 0x01},
   {"mode", 0x02},
   {NULL, 0}
};

/* "client" zone variables */
static struct cfg_Var_inZone cZ_client_vlist[] = {
   {"nickname", 0x01},
   {"username", 0x02},
   {"hostname", 0x04},
   {"realname", 0x08},
   {"ip", 0x10},
   {"mode", 0x20},
   {"join", 0x40},
   {"enabled", 0x80},
   {NULL, 0}
};

/* "perform" zone variables */
static struct cfg_Var_inZone cZ_perform_vlist[] = {
   {"scanthreads", 0x01},
   {"recheck", 0x02},
   {"recheckshared", 0x04},
   {"timeout_connect", 0x08},
   {"timeout_send", 0x10},
   {"timeout_listen", 0x20},
   {"cache", 0x40},
   {"http_ports", 0x80},
   {NULL, 0}
};

/* "share" zone variables */
static struct cfg_Var_inZone cZ_share_vlist[] = {
   {"name", 0x01},
   {"successor", 0x02},
   {"server", 0x04},
   {NULL, 0}
};

/* "log" zone variables */
static struct cfg_Var_inZone cZ_log_vlist[] = {
   {"directory", 0x0001},
   {"current", 0x0002},
   {"daily", 0x0004},
   {"update", 0x0008},
   {"history", 0x0010},
   {"system", 0x0020},
   {"glines", 0x0040},
   {"privmsg", 0x0080},
   {"opercmds", 0x0100},
   {"listener", 0x0200},
   {"stats", 0x0400},
   {NULL, 0}
};

/* "gline" zone variables */
static struct cfg_Var_inZone cZ_gline_vlist[] = {
	{"delay", 0x0001},
	{"cached", 0x0002},
	{"direct", 0x0004},
	{"reason_wingate", 0x0008},
	{"reason_socks", 0x0010},
	{"reason_http", 0x0020},
	{"relay", 0x0040},
	{NULL, 0}
};

static char *nolist[] = {
	"no", "n", "0", "false", "off"
};

static unsigned int checkVar_inZone(char *v, unsigned int z);
static Boolean checkMissing(char *zone, unsigned int z, unsigned int opt);
static Boolean retrieve_http_ports(char *ar);
static void test_vhost(unsigned long inAddr);

inline void
syntaxErr(int n)
{
   fprintf(stderr, "Configuration file: syntax error (line %d).\n", n);
}

/* This only checks for a legal aspect of an IP,
 * to avoid inet_aton on very strange things... */
inline Boolean
IsLegalIP(const char *ip)
{
   int len = strlen(ip) ;
   if(len >= 16)
      return false ;
   return (strspn(ip, ".0123456789") == len);
}

inline void
tab2space(char *ioStr)
{
   char *p = ioStr;
   while(p && *p) {
      if(*p == '\t')
         *p = ' ';
      ++p;
   }
}

static Boolean
isTrue(const char *s)
{
   int i;
   for(i=0;i<sizeof(nolist)/sizeof(char *);i++)
      if(!stricmp(s, nolist[i]))
         return false;
   return true;
}

Boolean
conf_load()
{
   FILE *fp;
   char *al, *ar;
   char ch;
   char buf[256];
   char zone[32];
   int zone_pos = 0;
   unsigned int confZone = confZone_nil;
   unsigned int confZoneDone = confZone_nil;
   unsigned int v_id = 0;
   int line_counter = 1;
   Boolean flag_enabled = true;
   Boolean flag_ignore = false;
   unsigned int zoneOpt;
   const char* missing_str = "Configuration file: Loading FAILED (Missing zone %s)\n";
   
   Uplink_s *  MyUplinkNew = NULL;
   Uplink_s *  MyUplinkPtr = NULL;
   Share_s *  MyShareNew = NULL;
   Share_s *  MySharePtr = NULL;
   Serverlist *  servlistPtr = NULL;
   Serverlist *  servlistScanPtr = NULL;
   IPrangelist *  IPrange = NULL;
   IPrangelist *  IPrangePtr = NULL;
   Target_s * MyTargetNew = NULL;
   Target_s * MyTargetPtr = NULL;
   Relay_s * relayp = NULL;
   Relay_s * relayp2 = NULL;
   
   /* Pre-Init */
   Me = NULL;
   MyUplink = NULL;
   MyConsole = NULL;
   MyClient = NULL;
   MyPerform = NULL;
   MyShare = NULL;
   MyNoScan = NULL;
   MyLog = NULL;
   MyTarget = NULL;
   totalTarget = 0; /* Added when writing rehash support */
   MyGline = NULL;

   /* Open configuration file */
   if((fp = fopen(PROXYSCAN_CONF_FILENAME, "r")) == NULL) {
      printf("read_conf> Cannot open %s!\n", PROXYSCAN_CONF_FILENAME);
      return false;
   }

   while((ch = (char) fgetc(fp)) != EOF) {
      if(flag_ignore && ((ch != 0x0D) && (ch != 0x0A))) {
         continue;
      }
      switch (ch) {

            /* blank trailing */
         case '\t':
         case ' ':
            continue;


            /* new line */
         case '\r':
         case '\n':
            line_counter++;
            flag_ignore = false;
            break;
            
         case '#':
            flag_ignore = true;
            break;
            
         case '{':
            checkZone(confZone_nil);

            /* Valid conf zone? */
            if(zone_pos == 0) {
               syntaxErr(line_counter);
            }
            zone[zone_pos++] = 0;
            if(!stricmp(zone, CONFZONE_WORD_ME)) {

               confZone = confZone_me;
               Me = (struct Me_s *) xmalloc(sizeof(struct Me_s));   /* One and only */
               memset(Me, 0, sizeof(struct Me_s));

            } else if(!stricmp(zone, CONFZONE_WORD_UPLINK)) {

               confZone = confZone_uplink;
               MyUplinkNew = (struct Uplink_s *) xmalloc(sizeof(struct Uplink_s));
               memset(MyUplinkNew, 0, sizeof(struct Uplink_s));

            } else if(!stricmp(zone, CONFZONE_WORD_CONSOLE)) {

               confZone = confZone_console;
               MyConsole =
                  (struct Console_s *) xmalloc(sizeof(struct Console_s));   /* One and only */

            } else if(!stricmp(zone, CONFZONE_WORD_CLIENT)) {

               confZone = confZone_client;
               MyClient =
                  (struct MyClient_s *) xmalloc(sizeof(struct MyClient_s));   /* One and only */

            } else if(!stricmp(zone, CONFZONE_WORD_PERFORM)) {

               confZone = confZone_perform;
               MyPerform =
                  (struct Perform_s *) xmalloc(sizeof(struct Perform_s));   /* One and only */
               memset(MyPerform, 0, sizeof(struct Perform_s));

            } else if(!stricmp(zone, CONFZONE_WORD_SHARE)) {

               confZone = confZone_share;
               MyShareNew = (struct Share_s *) xmalloc(sizeof(struct Share_s));
               memset(MyShareNew, 0, sizeof(struct Share_s));  /* servlist == NULL */

            } else if(!stricmp(zone, CONFZONE_WORD_NOSCAN)) {

               confZone = confZone_noscan;
               MyNoScan = (struct NoScan_s *) xmalloc(sizeof(struct NoScan_s));   /* One and only */
               memset(MyNoScan, 0, sizeof(struct NoScan_s));
            
            } else if(!stricmp(zone, CONFZONE_WORD_LOG)) {
               
               confZone = confZone_log;
               MyLog = (struct Log_s *) xmalloc(sizeof(struct Log_s));
               memset(MyLog, 0, sizeof(struct Log_s));
            
            } else if(!stricmp(zone, CONFZONE_WORD_TARGET)) {
            
              confZone = confZone_target;
            
            } else if(!stricmp(zone, CONFZONE_WORD_GLINE)) {
              
              confZone = confZone_gline;
              MyGline = (struct Gline_s *) xmalloc(sizeof(struct Gline_s));
              memset(MyGline, 0, sizeof(struct Gline_s));
              
            } else {
               printf
                  ("Configuration file: Error: Unrecognized zone '%s' (line %d)\n",
                   zone, line_counter);
               fclose(fp);
               return false;
               
            }

          /*  printf("Entering Zone: %s\n", zone);  */

            confZoneDone |= confZone;
            flag_enabled = true;   /* Default state: enabled (for options-zone which can be disabled) */
            zoneOpt = 0;
            
            memset(buf, 0, sizeof(buf));
            
            /* Parsing zone */
            while(fgets(buf, sizeof(buf) - 1, fp)) {
               line_counter++;
               al = buf;
               if(*al == '#')
                  continue;
               stripline(al);   /* cleaning */
               trim_left(&al);  /* cleaning */
               tab2space(al);   /* cleaning */
               if(strlen(al) > 0) {
                  
                  /* End of Zone ? */
                  if(al && (*al == '}')) {
                     if(*(al + 1) == '\0') {
                        
                        /* Checking for missing fields */
                        if (flag_enabled && !checkMissing(zone, confZone, zoneOpt)) {
                           printf("Configuration file: Something's wrong (zone %s, line %d)\n",
                              zone, line_counter);
                           fclose(fp);
                           return false;
                        }
                        
                        /* Linking zone list */
                        if (flag_enabled && (confZone == confZone_uplink)) {
                           if (NULL == MyUplink) {
                              MyUplink = MyUplinkNew;
                           } else {
                              MyUplinkPtr = MyUplink;
                              while(MyUplinkPtr->next != NULL)
                                 MyUplinkPtr = MyUplinkPtr->next;
                              MyUplinkPtr->next = MyUplinkNew;   /* added at the end of the list */
                           }
                        
                        } else if (confZone == confZone_share) {
                           if (NULL == MyShare) {
                              MyShare = MyShareNew;
                           } else {
                              MySharePtr = MyShare;
                              while(NULL != MySharePtr->next)
                                 MySharePtr = MySharePtr->next;
                              MySharePtr->next = MyShareNew;   /* added at the end of the list */
                           }
                        
                        }
                        
                        
                        
                        confZone = confZone_nil;
                        zone_pos = 0;
                        zone[zone_pos] = '\0';
                        
                        break;
                     } else {
                        /* No - error */
                        printf
                           ("Configuration file: Parse error (zone %s, line %d)\n",
                            zone, line_counter);
                        fclose(fp);
                        return false;
                     }
                  }


                  /* noscan field has a different struct */
                  if (!(confZone & (confZone_noscan | confZone_target))) {
                     
                     al = strtok(al, "=");      /* left arg */
                     ar = strtok(NULL, "");     /* right arg */

                     trim_left(&al);
                     trim_right(al);

              /*      printf("parsing al= '%s'\n", al);  */

                     /* Missing field? */

                     if (!ar || !al) {
                        printf
                           ("Configuration file: Syntax error or missing arg (zone %s; line %d).\n",
                            zone, line_counter);
                        fclose(fp);
                        return false;
                     }

                     /* Checking variable */
                     v_id = checkVar_inZone(al, confZone);

                     if (v_id == 0) {
                        fclose(fp);
                        printf("Configuration file: Unknown variable '%s' (zone %s, line %d).\n",
                            al, zone, line_counter);
                        return false;
                     }
                     
                     /* Trim right arg */
                     trim_left(&ar);
                     if (confZone != confZone_gline)
                        trim_right(ar);
                     
                     if (confZone == confZone_me) {
                           
                           checkDouble(zoneOpt, v_id);
                           
                           switch (v_id) {
                           
                              case 0x01:   /* name */
                                 strncpy(Me->name, ar, MAXSERVERNAME_LENGTH - 1);
                                 break;
                                 
                              case 0x02:   /* numeric */
                                 Me->numeric = atoi(ar);
                                 if(Me->numeric <= 0) {
                                 	if(strcmp(ar, "0")) {    /* *Explicit* numeric 0 allowed. */
                                 		printf("Configuration file: Bad value for 'numeric' (zone %s, line %d).\n",
                                           zone, line_counter);
                                        fclose(fp);
                                        return false;
                                 	}
                                 }
#ifdef N2K_SUPPORT
                                 else if(Me->numeric >= 4096) {
                                    printf("Configuration file: Bad value for (extended) 'numeric' (zone %s, line %d).\n",
                                          zone, line_counter);
                                    fclose(fp);
                                    return false;
                                 }
#else
                                 else if(Me->numeric >= 64) {
                                    printf("Configuration file: Bad value for (not extended) 'numeric' (zone %s, line %d).\n",
                                          zone, line_counter);
                                    fclose(fp);
                                    return false;
                                 }
#endif
                                 break;
                                 
                              case 0x04:   /* capacity */
                                 Me->capacity = atoi(ar);
                                 if(Me->capacity <= 0) {
                                 	printf("Configuration file: Bad value for 'maxclients' (zone %s, line %d).\n",
                                         zone, line_counter);
                                    fclose(fp);
                                    return false;
                                 }
#ifdef N2K_SUPPORT
                                 else if(Me->capacity > 2621144) {
                                    printf("Configuration file: Bad value for (extended) 'capacity' (zone %s, line %d).\n",
                                          zone, line_counter);
                                    fclose(fp);
                                    return false;
                                 }
#else
                                 else if(Me->capacity > 4096) {
                                    printf("Configuration file: Bad value for (not extended) 'capacity' (zone %s, line %d).\n",
                                          zone, line_counter);
                                    fclose(fp);
                                    return false;
                                 }
#endif
                                 break;
                                 
                              case 0x08:   /* serverinfo */
                                 strncpy(Me->serverinfo, ar, MAXSERVERINFO_LENGTH - 1);
                                 break;
                                 
                              case 0x10:   /* hostip */
                                 strncpy(Me->hostip, ar, 16 - 1);
                                 break;
                                 
                              case 0x20:   /* bindip */
                                 strncpy(Me->bindip, ar, 16 - 1);
                                 /* Validity checked here to avoid it in each scan thread */
                                 if (strlen(Me->bindip) >= 7)
                                    Me->vhost = true;
                                 break;
                              
                              case 0x40:   /* listener */
                                 strncpy(Me->listener, ar, 16 - 1);
                                 break;
                                 
                              case 0x80:   /* id */
                                 if (strlen(ar) > 9) {
                                    printf("Configuration file: 'id too long (zone %s, line %d).\n",
                                         zone, line_counter);
                                    return false;
                                 }
                                 strncpy(Me->id, ar, 10 - 1);
                                 break;
                                
                              default:
                              	Debug((DBLMAX, "PARSE ERROR DEFAULT conf.c"));
                              	break;
                           }
                           
                       } else if (confZone == confZone_uplink) {
                          
                          if (!flag_enabled) {
                             continue;
                          }
                          
                          checkDouble(zoneOpt, v_id);
                          
                          switch(v_id) {
                          
                             case 0x01:   /* host */
                                strncpy(MyUplinkNew->host, ar, MAXSERVERNAME_LENGTH - 1);
                                break;
                             
                             case 0x02:   /* bindip */
                                strncpy(MyUplinkNew->bindip, ar, 16 - 1);
                                if(strlen(MyUplinkNew->bindip) >= 7)
                                   MyUplinkNew->vhost = true;
                                break;
                             
                             case 0x04:   /* port */
                                MyUplinkNew->port = (unsigned short) atoi(ar);
                                if(MyUplinkNew->port == 0 || MyUplinkNew->port > 65535) {
                                   printf("Configuration file: Bad value for 'port' (zone %s, line %d).\n",
                                         zone, line_counter);
                                   fclose(fp);
                                   return false;
                                }
                                break;
                                
                             case 0x08:   /* password */
                                strncpy(MyUplinkNew->password, ar, MAXPASSWORD_LENGTH);
                                break;
                                
                             case 0x10:   /* enabled */
                                flag_enabled = isTrue(ar);
                                break;
                                
                             case 0x20:   /* priority */
                                MyUplinkNew->priority = atoi(ar);
                                if(MyUplinkNew->priority == 0) {
                                   printf("Configuration file: Bad value for 'priority' (zone %s, line %d).\n",
                                         zone, line_counter);
                                   return false;
                                }
                                break;
                          
                           }
                           
                           
                       
                       } else if(confZone == confZone_console) {
                          
                          checkDouble(zoneOpt, v_id);
                          
                          switch (v_id) {
                          
                             case 0x01:   /* channel */
                                if(strlen(ar) > 64) {
                                   printf("Configuration file: Channel name too long.. (zone %s, line %d).\n",
                                         zone, line_counter);
                                   return false;
                                }
                                MyConsole->channel = (char *) xmalloc(sizeof(char) * (strlen(ar) + 1));
                                strcpy(MyConsole->channel, ar);
                                break;
                                
                              case 0x02:
                                 if((strlen(ar) > 14) || (*ar != '+')) {
                                   syntaxErr(line_counter);
                                   fclose(fp);
                                   return false;
                                 }
                                 strncpy(MyConsole->mode, ar, 14);
                                 break;
                          }
                       
                       } else if (confZone == confZone_client) {
                          
                          checkDouble(zoneOpt, v_id);
                          
                          if (!flag_enabled) {
                             continue;
                          }
                          
                          switch(v_id) {
                          
                             case 0x01:   /* nickname */
                                strncpy(MyClient->nickname, ar, MAXNICKNAME_LENGTH - 1);
                                break;
                                
                             case 0x02:   /* username */
                                strncpy(MyClient->username, ar, MAXUSERNAME_LENGTH - 1);
                                break;
                                
                             case 0x04:   /* hostname */
                                MyClient->hostname = (char *) xmalloc(sizeof(char) * (strlen(ar) + 1));
                                strcpy(MyClient->hostname, ar);
                                break;
                                
                             case 0x08:   /* realname */
                                MyClient->realname = (char *) xmalloc(sizeof(char) * (strlen(ar) + 1));
                                strcpy(MyClient->realname, ar);
                                break;
                             
                             case 0x10:   /* ip */
			        MyClient->ip.s_addr = inet_addr(ar);
                                if(!IsLegalIP(ar) || (strcmp(ar, "255.255.255.255") && (MyClient->ip.s_addr == INADDR_NONE))) {
                                   syntaxErr(line_counter);
                                   fclose(fp);
                                   return false;
                                }
                                MyClient->ip.s_addr = ntohl(MyClient->ip.s_addr);
                                break;
                             
                             case 0x20:   /* mode */
                                if((strlen(ar) > 14) || (*ar != '+')) {
                                   syntaxErr(line_counter);
                                   fclose(fp);
                                   return false;
                                }
                                strncpy(MyClient->mode, ar, 14);
                                break;
                                
                             case 0x40:   /* join */
                                MyClient->join = isTrue(ar);
                                break;
                                 
                             case 0x80:   /* enabled */
                                flag_enabled = isTrue(ar);
                                if(!flag_enabled) {
	                                MyClient->nickname[ 0 ] = 0;
    	                            MyClient->username[ 0 ] = 0;
        	                        if(NULL != MyClient->hostname) {
            	                    	xfree(MyClient->hostname);
                	                	MyClient->hostname = NULL;
                    	            }
                    	            MyClient->realname = NULL;
                        	        MyClient->join = 0;
                        	    }
                                break;
                          }
                       
                       } else if (confZone == confZone_perform) {
                          
                          checkDouble(zoneOpt, v_id);
                          
                          switch(v_id) {
                          
                             case 0x01:   /* scanthreads */
                                MyPerform->scanthreads = (long) atoi(ar);
                                if(MyPerform->scanthreads <= 0) {
                                	printf("Configuration file: Bad value for 'scanthreads' (zone %s, line %d).\n",
                                        zone, line_counter);
                                    fclose(fp);
                                    return false;
                                }
                                break;
                             
                             case 0x02:   /* recheck */
                                MyPerform->recheck = (long) atoi(ar);
                                if(MyPerform->recheck <= 5) {
                                	printf("Configuration file: Bad value for 'recheck' (zone %s, line %d).\n",
                                        zone, line_counter);
                                    fclose(fp);
                                    return false;
                                }
                                break;
                             
                             case 0x04:   /* recheckshared */
                                MyPerform->recheckshared = (long) atoi(ar);
                                if(MyPerform->recheck <= 5) {
                                	printf("Configuration file: Bad value for 'recheckshared' (zone %s, line %d).\n",
                                        zone, line_counter);
                                    fclose(fp);
                                    return false;
                                }
                                break;
                                
                             case 0x08:   /* timeout_connect */
                                MyPerform->timeout_connect = (long) atoi(ar);
                                if((MyPerform->timeout_connect <= 0) || (MyPerform->timeout_connect >= 300)) {
                                	printf("Configuration file: Bad value for 'timeout_connect' (zone %s, line %d).\n",
                                        zone, line_counter);
                                    fclose(fp);
                                    return false;
                                }
                                break;
                                
                             case 0x10:   /* timeout_send */
                                MyPerform->timeout_send = (long) atoi(ar);
                                if((MyPerform->timeout_send <= 0) || (MyPerform->timeout_send >= 300)) {
                                	printf("Configuration file: Bad value for 'timeout_send' (zone %s, line %d).\n",
                                        zone, line_counter);
                                    fclose(fp);
                                    return false;
                                }
                                break;
                                
                             case 0x20:   /* timeout_listen */
                                MyPerform->timeout_listen = (long) atoi(ar);
                                if((MyPerform->timeout_listen <= 0) || (MyPerform->timeout_listen >= 300)) {
                                	printf("Configuration file: Bad value for 'timeout_listen' (zone %s, line %d).\n",
                                        zone, line_counter);
                                    fclose(fp);
                                    return false;
                                }
                                break;
                                
                             case 0x40:   /* cache */
                                MyPerform->cachesize = (long) atoi(ar);
                                if((MyPerform->cachesize <= 0) || (MyPerform->cachesize > 300000000L)) { /* 100+ MB */
                                	printf("Configuration file: Bad value for 'cache' (zone %s, line %d).\n",
                                        zone, line_counter);
                                    fclose(fp);
                                    return false;
                                }
                                break;
                                
                              case 0x80:   /* http_ports */
                                if (!retrieve_http_ports(ar)) {
                                   printf("Configuration file: Bad value for 'http_ports' (zone %s, line %d).\n",
                                        zone, line_counter);
                                   fclose(fp);
                                    return false;
                                }
                                break;
                          }
                                
                       } else if (confZone == confZone_share) {
                       
                          switch (v_id) {
                          
                             case 0x01:   /* name */
                                checkDouble(zoneOpt, v_id);
                                strnlwrcpy(MyShareNew->name, ar, MAXSERVERNAME_LENGTH - 1);
                                break;
                             
                             case 0x02:   /* successor */
                                checkDouble(zoneOpt, v_id);
                                strnlwrcpy(MyShareNew->successor, ar, MAXSERVERNAME_LENGTH - 1);
                                break;
                                
                             case 0x04:   /* server --several server masks supported */
                                servlistPtr = (struct Serverlist *) xmalloc(sizeof(struct Serverlist));
                                strnlwrcpy(servlistPtr->name, ar, MAXSERVERNAME_LENGTH - 1);
                                servlistPtr->next = NULL;
                                /* link */
                                
                                if(NULL == MyShareNew->servlist) {
                                   MyShareNew->servlist = servlistPtr;
                                } else {
                                   servlistScanPtr = MyShareNew->servlist;
                                   while(NULL != servlistScanPtr->next)
                                      servlistScanPtr = servlistScanPtr->next;
                                   servlistScanPtr->next = servlistPtr;
                                }
                                break;
                                
                          }
                          
                       
                       
                       } else if (confZone == confZone_log) {
                       
                           switch (v_id) {
                              
                              case 0x0001:   /* current */
                                 if(*(ar + strlen(ar) - 1) == '/') {
                                    printf("Configuration file: While parsing 'directory': Unneeded '/' character (zone %s, line %d).\n",
                                        zone, line_counter);
                                    fclose(fp);
                                    return false;
                                 }
                                 strncpy(MyLog->dir, ar, MAXFILENAME_LENGTH - 1);
                                 break;
                                 
                              case 0x0002:   /* current */
                                 strncpy(MyLog->current, ar, MAXFILENAME_LENGTH - 1);
                                 break;
                              
                              case 0x0004:   /* daily */
                                 strncpy(MyLog->daily, ar, MAXFILENAME_LENGTH - 1);
                                 break;
                              
                              case 0x0008:   /* update */
                                 MyLog->update = atoi(ar);
                                 if((MyLog->update < 0) || (MyLog->update > 23)) {
                                    printf("Configuration file: Bad value for 'update' (zone %s, line %d).\n",
                                        zone, line_counter);
                                    fclose(fp);
                                    return false;
                                 }
                                 break;
                              
                              case 0x0010:   /* history */
                                 strncpy(MyLog->history_dir, ar, MAXFILENAME_LENGTH - 1);
                                 break;
                              
                              case 0x0020:   /* pxys system */
                                 strncpy(MyLog->pxys, ar, MAXFILENAME_LENGTH - 1);
                                 break;
                              
                              case 0x0040:   /* glines */
                                 strncpy(MyLog->glines, ar, MAXFILENAME_LENGTH - 1);
                                 break;
                              
                              case 0x0080:   /* privmsg */
                                 strncpy(MyLog->privmsg, ar, MAXFILENAME_LENGTH - 1);
                                 break;
                                 
                              case 0x0100:   /* opercmds */
                                 strncpy(MyLog->opercmds, ar, MAXFILENAME_LENGTH - 1);
                                 break;
                              
                              case 0x0200:   /* listener */
                                 strncpy(MyLog->listener, ar, MAXFILENAME_LENGTH - 1);
                                 break;
                                 
                              case 0x0400:   /* stats */
                                 strncpy(MyLog->stats, ar, MAXFILENAME_LENGTH - 1);
                                 break;
                           }
                           
                       
                       } else if (confZone == confZone_gline) {
                          
                          switch (v_id) {
                              
                              case 0x0001:   /* delay */
                                 MyGline->delay = atoi(ar);
                                 if(MyGline->delay <= 0) {
                                    printf("Configuration file: Bad value for 'delay' (%ld) (zone %s, line %d).\n",
                                        MyGline->delay, zone, line_counter);
                                    fclose(fp);
                                    return false;
                                 }
                                 break;
                              case 0x0002:   /* cached */
                                 MyGline->cached = atoi(ar);
                                 if(MyGline->cached <= 0) {
                                    printf("Configuration file: Bad value for 'cached' (%ld) (zone %s, line %d).\n",
                                        MyGline->cached, zone, line_counter);
                                    fclose(fp);
                                    return false;
                                 }
                                 break;
                              case 0x0004:   /* direct */
                                 MyGline->direct = isTrue(ar);
                                 break;
#ifdef SCAN_WINGATE
                              case 0x0008:   /* reason_wingate */
                                 strncpy(MyGline->reason_wingate, ar,
                                    sizeof(MyGline->reason_wingate) - 1);
                                 break;
#endif
#ifdef SCAN_SOCKS
                              case 0x0010:   /* reason_socks */
                                 strncpy(MyGline->reason_socks, ar,
                                    sizeof(MyGline->reason_socks) - 1);
                                 break;
#endif
#ifdef SCAN_HTTP
                              case 0x0020:   /* reason_http */
                                 strncpy(MyGline->reason_http, ar,
                                    sizeof(MyGline->reason_http) - 1);
                                 break;
#endif
                              case 0x0040:   /* relay */
                                 relayp = (struct Relay_s *) xmalloc(sizeof(struct Relay_s));
                                 al = strtok(ar, " ");
                                 strncpy(relayp->nick, al, sizeof(relayp->nick) - 1);
                                 ar = strtok(NULL, " ");
                                 if(ar && !stricmp(ar, "on"))
                                     ar = strtok(NULL, " ");
                                 if(NULL == ar) {
                                    printf("Configuration file: Bad value for 'relay' (zone %s, line %d).\n",
                                        zone, line_counter);
                                    fclose(fp);
                                    return false;
                                 }
                                 strnlwrcpy(relayp->serv, ar, sizeof(relayp->serv) - 1);
                                 
                                 relayp->next = NULL;
                                 
                                 if(NULL == MyGline->relay) {
                                    MyGline->relay = relayp;
                                 } else {
                                    relayp2 = MyGline->relay;
                                    while(relayp2->next)
                                       relayp2 = relayp2->next;
                                    relayp2->next = relayp;
                                 }

				 /* auth ? */
                                 ar = strtok(NULL, " ");
                                 if(NULL != ar) {
                                    if(stricmp(ar, "auth")) {
                                       printf("Configuration file: Bad value for 'relay' (auth) (zone %s, line %d).\n",
                                            zone, line_counter);
                                       fclose(fp);
                                       return false;
                                    }
                                    ar = strtok(NULL, "");
                                    if(NULL == ar) {
                                       printf("Configuration file: Bad value for 'relay' (auth) (zone %s, line %d).\n",
                                            zone, line_counter);
                                       fclose(fp);
                                       return false;
                                    }
                                    if(strlen(ar) >= 64) {
                                       printf("Configuration file: Bad value for 'relay' (too long) (zone %s, line %d).\n",
                                            zone, line_counter);
                                       fclose(fp);
                                       return false;
                                    }
                                    strncpy(relayp->auth, ar, sizeof(relayp->auth));
                                 }

                                 break;
                          }                              
                              
                       } else {
                          Debug((DBLMAX, "conf.c: BUG"));
                       }
                   
                       
                  } else if (confZone == confZone_noscan) {
                     /* noscan field */
                     
                     /* noscan options begins with: - IP               fixed IP
                      *                             - 'from' IP 'to'   range of IPs
                      *                             - 'on'             an IRC server (mask) or our network
                      */
                      
                      
                      al = strtok(al, " ");
                      if (NULL == al)
                         break;
                      
                      if(!stricmp(al, "from")) {
                         
                         
                         
                         al = strtok(NULL, " ");
                         if (NULL == al)
                            syntaxErr(line_counter);
                            
                         ar = strtok(NULL, " ");
                         if (NULL == al)
                            syntaxErr(line_counter);
                         
                         if (!stricmp(ar, "to")) {
                            ar = strtok(NULL, " ");
                            if (NULL == al)
                               syntaxErr(line_counter);
                         }
                         
                         /* IPs range */
                         IPrangePtr = (struct IPrangelist *) xmalloc(sizeof(struct IPrangelist));
                         IPrangePtr->next = NULL;
                         
                         IPrangePtr->start.s_addr = inet_addr(al);
                         if(!IsLegalIP(al) || (strcmp(al, "255.255.255.255") && (IPrangePtr->start.s_addr == INADDR_NONE))) {
                            printf("Configuration file: Bogus IP (from) (line %d)\n", line_counter);
                            fclose(fp);
                            return false;
                         }
                         
                         IPrangePtr->stop.s_addr = inet_addr(ar);
                         if (!IsLegalIP(ar) || (strcmp(ar, "255.255.255.255") && (IPrangePtr->stop.s_addr == INADDR_NONE))) {
                            printf("Configuration file: Bogus IP (to) (line %d)\n", line_counter);
                            fclose(fp);
                            return false;
                         }
                         
                         /* Reversing bytes order */
                         IPrangePtr->start.s_addr = ntohl(IPrangePtr->start.s_addr);
                         IPrangePtr->stop.s_addr = ntohl(IPrangePtr->stop.s_addr);
                         
                         /* Checking if ipmin <= ipmax */
                         if(IPrangePtr->start.s_addr > IPrangePtr->stop.s_addr) {
                            printf("Configuration file: Error line %d\n", line_counter);
                            printf("Configuration file: Syntax: from <IPmin> to <IPmax>\n");
                            fclose(fp);
                            return false;
                         }
                         
                         if(NULL == MyNoScan->ips) {
                            MyNoScan->ips = IPrangePtr;
                         } else {
                            IPrange = MyNoScan->ips;
                            while(NULL != IPrange->next)
                               IPrange = IPrange->next;
                            IPrange->next = IPrangePtr;
                         }
                         
                      } else if (!stricmp(al, "on")) {
                         /* IRC Server */
                         
                         ar = strtok(NULL, " ");
                         if(NULL == ar)
                            syntaxErr(line_counter);
                         
                         servlistPtr = (struct Serverlist *) xmalloc(sizeof(struct Serverlist));
                         strncpy(servlistPtr->name, ar, MAXSERVERNAME_LENGTH - 1);
                         servlistPtr->next = NULL;
                         
                         
                         if(NULL == MyNoScan->residences) {
                            MyNoScan->residences = servlistPtr;
                         } else {
                            servlistScanPtr = MyNoScan->residences;
                            while(NULL != servlistScanPtr->next)
                               servlistScanPtr = servlistScanPtr->next;
                            servlistScanPtr->next = servlistPtr;
                         }
                         
                         
                      } else if (IsLegalIP(al)) {
                         /* Single IP */
                         
                         IPrangePtr = (struct IPrangelist *) xmalloc(sizeof(struct IPrangelist));
                         IPrangePtr->next = NULL;
                         
                         IPrangePtr->start.s_addr = inet_addr(al);
                         /* We need to be solaris compliant here...
                          * if(!inet_aton(al, &IPrangePtr->start))
                          */
                         if ((strcmp(al, "255.255.255.255") && (IPrangePtr->start.s_addr == INADDR_NONE))) {
                            printf("Configuration file: Bogus IP %s (line %d)\n", al, line_counter);
                            fclose(fp);
                            return false;
                         }
                         
                         /* Reversing bytes order */
                         IPrangePtr->start.s_addr = ntohl(IPrangePtr->start.s_addr);
                         
                         
                         /* Single IP - from <singleIP> to <singleIP> will do that. */
                         memcpy(&IPrangePtr->stop, &IPrangePtr->start, sizeof(struct in_addr));
                         
                         if (NULL == MyNoScan->ips) {
                            MyNoScan->ips = IPrangePtr;
                         } else {
                            IPrange = MyNoScan->ips;
                            while(NULL != IPrange->next)
                               IPrange = IPrange->next;
                            IPrange->next = IPrangePtr;
                         }
                         
                      } else {
                         syntaxErr(line_counter);
                         fclose(fp);
                         return false;
                      }
                      

                  } else {
                     /* target zone (socks4 target host) */
                     
                     al = strtok(al, " ");
                      if (!al)
                         break;
                     
                     if (strlen(al) <= 4) {
                        printf("Zone %s: ignoring host: %s\n", zone, al);
                     } else {
                        MyTargetNew = (struct Target_s *) xmalloc(sizeof(struct Target_s));
                        memset(MyTargetNew, 0, sizeof(struct Target_s));
                        strncpy(MyTargetNew->name, al, MAXSERVERNAME_LENGTH - 1);
                        MyTargetNew->next = NULL;
                        if (NULL == MyTarget) {
                           MyTarget = MyTargetNew;
                        } else {
                           MyTargetPtr = MyTarget;
                           while(NULL != MyTargetPtr->next)
                              MyTargetPtr = MyTargetPtr->next;
                           MyTargetPtr->next = MyTargetNew;
                        }
                        totalTarget++;
                     }
                  }
                  
                  if (!(zoneOpt & v_id))
                        zoneOpt |= v_id;
               }


            }
            break;

         case '}':
            fclose(fp);
            printf("Configuration file: Parse error (line %d)\n", line_counter);
            return false;

         default:
            if (!isspace((int)ch)) {
               zone[zone_pos++] = ch;
            }
      }
   }
   
   fclose(fp);
   
   if(!(confZoneDone & confZone_me)) {
      printf(missing_str, CONFZONE_WORD_ME);
      return false;
   }
   if(!(confZoneDone & confZone_uplink)) {
      printf(missing_str, CONFZONE_WORD_UPLINK);
      return false;
   }
   if(!(confZoneDone & confZone_console)) {
      printf(missing_str, CONFZONE_WORD_CONSOLE);
      return false;
   }
   if(!(confZoneDone & confZone_client)) {
      printf(missing_str, CONFZONE_WORD_CLIENT);
      return false;
   }
   if(!(confZoneDone & confZone_perform)) {
      printf(missing_str, CONFZONE_WORD_PERFORM);
      return false;
   }
   /* !allow no share zone!
   if(!(confZoneDone & confZone_share)) {
      printf(missing_str, CONFZONE_WORD_SHARE);
      return false;
   }
   */
   
   if(!(confZoneDone & confZone_noscan)) {
      printf(missing_str, CONFZONE_WORD_NOSCAN);
      return false;
   }
   if(!(confZoneDone & confZone_log)) {
      printf(missing_str, CONFZONE_WORD_LOG);
      return false;
   }
   if(!(confZoneDone & confZone_target)) {
      printf(missing_str, CONFZONE_WORD_TARGET);
      return false;
   }
   if(!(confZoneDone & confZone_gline)) {
      printf(missing_str, CONFZONE_WORD_GLINE);
      return false;
   }
   
   
#if 0
   /***Recap***/
   
   /* Me */
   printf("* zone Me:\n");
   printf("* Me->name = %s\n", Me->name);
   printf("* Me->numeric = %ld\n", Me->numeric);
   printf("* Me->capacity = %ld\n", Me->capacity);
   printf("* Me->serverinfo = %s\n", Me->serverinfo);
   printf("* Me->hostip = %s\n", Me->hostip);
   printf("* Me->id = %s\n", Me->id);
   printf("* \n");
   
   /* UplinkS */
   MyUplinkPtr = MyUplink;
   while(MyUplinkPtr != NULL) {
		printf("* zone Uplink:\n");
		printf("* MyUplink->host = %s\n", MyUplinkPtr->host);
		printf("* MyUplink->port = %u\n", MyUplinkPtr->port);
		printf("* MyUplink->password = %s\n", MyUplinkPtr->password);
		printf("* MyUplink->priority = %d\n", MyUplinkPtr->priority);
		printf("* \n");
		MyUplinkPtr = MyUplinkPtr->next;
   }
   
   /* Console */
   printf("* zone Console:\n");
   printf("* MyConsole->channel = %s\n", MyConsole->channel);
   printf("* \n");
   
   /* Client */
   printf("* zone Client:\n");
   if(MyClient->hostname == NULL) {
      printf("* MyClient disabled.\n");
   } else {
      printf("* MyClient->nickname = %s\n", MyClient->nickname);
      printf("* MyClient->username = %s\n", MyClient->username);
      printf("* MyClient->realname = %s\n", MyClient->realname);
      printf("* MyClient->join = %d\n", MyClient->join); 
   }
   printf("* \n");
   
   /* Perform */
   printf("* zone Perform:\n");
   printf("* MyPerform->scanthreads = %ld\n", MyPerform->scanthreads);
   printf("* MyPerform->recheck = %ld\n", MyPerform->recheck);
   printf("* MyPerform->timeout_connect = %ld\n", MyPerform->timeout_connect);
   printf("* MyPerform->timeout_send = %ld\n", MyPerform->timeout_send);
   printf("* MyPerform->timeout_listen = %ld\n", MyPerform->timeout_listen);
   printf("* MyPerform->cache = %ld\n", MyPerform->cachesize);
   printf("* \n");
   
   /* ShareS */
   MySharePtr = MyShare;
   while(MySharePtr != NULL) {
      struct Serverlist *sl;
      printf("* zone Share:\n");
      printf("* MyShare->name = %s\n", MySharePtr->name);
      printf("* MyShare->successor = %s\n", MySharePtr->successor);
      
      /* servlist */
      sl = MySharePtr->servlist;
      while(sl != NULL) {
         printf("* MyShare->servlist->name = %s\n", sl->name);
         sl = sl->next;
      }
      printf("* \n");
      MySharePtr = MySharePtr->next;
   }
   
   /* NoScan */
   printf("* zone NoScan:\n");
   IPrangePtr = MyNoScan->ips;
   while(IPrangePtr != NULL) {
      char ip[20];
      al = inet_ntoa((const char *) &IPrangePtr->start);
      strcpy(ip, al);
      ar = inet_ntoa((const char *) &IPrangePtr->stop);
      printf("* NoScan->ips = %s (%lu) - %s (%lu)\n",
         ip, (unsigned long) IPrangePtr->start.s_addr, ar, (unsigned long) IPrangePtr->stop.s_addr);
      IPrangePtr = IPrangePtr->next;
   }
   
   servlistPtr = MyNoScan->residences;
   while(servlistPtr != NULL) {
      printf("* NoScan->residences = %s\n", servlistPtr->name);
      servlistPtr = servlistPtr->next;
   }
   sleep(2);
#endif
   
   
   return true;
}


static unsigned int
checkVar_inZone(char *v, unsigned int z)
{
   unsigned int ret = 0;
   struct cfg_Var_inZone *vz;
   int i = 0;

   if(z == confZone_me)
      vz = cZ_me_vlist;
   else if(z == confZone_uplink)
      vz = cZ_uplink_vlist;
   else if(z == confZone_console)
      vz = cZ_console_vlist;
   else if(z == confZone_client)
      vz = cZ_client_vlist;
   else if(z == confZone_perform)
      vz = cZ_perform_vlist;
   else if(z == confZone_share)
      vz = cZ_share_vlist;
   else if(z == confZone_log)
      vz = cZ_log_vlist;
   else if(z == confZone_gline)
      vz = cZ_gline_vlist;
   else
      return ret;
   
   do {
      if(stricmp(v, vz[i].vStr) == 0) {
         ret = vz[i].vID;
         break;
      }
   } while(vz[++i].vStr != NULL);

   return ret;
}

static Boolean
checkMissing(char *zone, unsigned int z, unsigned int opt)
{
   struct cfg_Var_inZone *vz;
   int i = 0;

   if(z == confZone_me)
      vz = cZ_me_vlist;
   else if(z == confZone_uplink)
      vz = cZ_uplink_vlist;
   else if(z == confZone_console)
      vz = cZ_console_vlist;
   else if(z == confZone_client)
      vz = cZ_client_vlist;
   else if(z == confZone_perform)
      vz = cZ_perform_vlist;
   else if(z == confZone_share)
      vz = cZ_share_vlist;
   else if(z == confZone_noscan)
      return true;
   else if(z == confZone_log)
      vz = cZ_log_vlist;
   else if(z == confZone_target)
      return true;
   else if(z == confZone_gline)
      return ((MyGline->delay > 0) && (MyGline->cached > 0)
#ifdef SCAN_WINGATE
         && (MyGline->reason_wingate[0] != 0)
#endif
#ifdef SCAN_SOCKS
         && (MyGline->reason_socks[0] != 0)
#endif
#ifdef SCAN_HTTP
         && (MyGline->reason_http != 0)
#endif
         );
   else {
      Debug((DBLMAX, "conf.c:checkMissing:Unknown zone."));
      return false;
   }
   
   do {
      if(!(opt & vz[i].vID)) {
         printf("Configuration file: Missing arg (%s) in zone %s.\n", vz[i].vStr, zone);
         return false;
      }
   } while(vz[++i].vStr != NULL);
   
   return true;
}


static Boolean
retrieve_http_ports(char *ar)
{
   int http_ports_count = 0;
   unsigned short *portp;
   char *p = ar, *q = NULL;
   int c;
   
   while ((c = *p++))
      if ( (c == ',' && isdigit(*p)) || (!http_ports_count && isdigit(c)) )
         http_ports_count++;
   
   if (http_ports_count > 0) {
      MyPerform->http_ports_count = http_ports_count;
      
      MyPerform->http_ports = (unsigned short *)
                       xmalloc(http_ports_count * sizeof(unsigned short)); /* LEAK? */
      portp = MyPerform->http_ports;
      for (p = ar; p; p = q, ++portp) {
         if ((q = strchr(p, ','))) {
            *q++ = '\0';
            *portp = (unsigned short)atoi(p);
            if (*portp == 0 || *portp > 65535)
               return false;
         } else {
            *portp = (unsigned short)atoi(p);
            if (*portp == 0 || *portp > 65535)
               return false;
            break; /* last port */
         }
      }
   }
   return true;
}

static void
test_vhost(unsigned long inAddr)
{
   struct sockaddr_in localAddr;
   int fd;
   
   memset(&localAddr, 0, sizeof(struct sockaddr_in));
   localAddr.sin_family = AF_INET;
   localAddr.sin_addr.s_addr = inAddr;
   localAddr.sin_port = htons(0);
   
   errno = 0;
   fd = socket(AF_INET, SOCK_STREAM, 0);
   if (fd < 0) {
      fprintf(stderr, "pxys: test_vhost> socket failed (errno=%d (%s))\n",
              errno, strerror(errno));
      exit(1);
   }
   
   errno = 0;
   if (bind(fd, (struct sockaddr *) &localAddr, sizeof(localAddr)) < 0) {
      fprintf(stderr, "pxys: test_vhost> bindfailed for %s (errno=%d (%s))\n",
              inet_ntoa(localAddr.sin_addr), errno, strerror(errno));
      exit(1);
   }
   
   close(fd);
}

int
conf_resolv()
{
   struct hostent *hp;
   struct Target_s *target;
   struct Uplink_s *uplink;
   int count = 0;
   
   if (Me->vhost) {
	  Me->bindaddr = inet_addr(Me->bindip);
      if(INADDR_NONE == Me->bindaddr) {
         fprintf(stderr, "pxys: virtual host resolution failed (me->bindip)\n");
         exit(1);
      }
      test_vhost(Me->bindaddr);
   }
   
   if (strcmp(Me->listener, "0")) {
      Me->listeneraddr = inet_addr(Me->listener);
      if(INADDR_NONE == Me->listeneraddr) {
         fprintf(stderr, "pxys: virtual host resolution failed (me->listener)\n");
         exit(1);
      }
   } else
      Me->listeneraddr = htonl(INADDR_ANY);
   
   test_vhost(Me->listeneraddr);
   
   for (uplink = MyUplink; uplink; uplink = uplink->next) {
      if(uplink->vhost) {
         uplink->bindaddr = inet_addr(uplink->bindip);
         if(INADDR_NONE == uplink->bindaddr) {
            fprintf(stderr, "pxys: virtual host resolution failed (uplink->bindip)\n");
            exit(1);
         }
         test_vhost( uplink->bindaddr);
         break;
      }
   }
   
   printf("pxys: resolving socks4 target hosts: ");
   fflush(stdout);
   
   for (target = MyTarget; target; target = target->next) {
      if ((hp = gethostbyname(target->name)) == NULL) {
         fprintf(stderr, "\nError: Can't resolve %s, aborting.", target->name);
         fflush(stdout);
         exit(1);
      } else {
         memcpy(&target->addr, hp->h_addr, hp->h_length);
         printf(".");
         fflush(stdout);
         Debug((DBLMAX, "Name server: hostname %s = %s", target->name, inet_ntoa(target->addr)));
         ++count;
      }
   }
   return count;
}


void
conf_delete()
{
   Uplink_s *uplink, *uplink_next;
   Share_s *share, *share_next;
   Target_s *tgt, *tgt_next;
   Serverlist *srv, *srv_next;
      
   if (Me) {
      xfree(Me);
      Me = NULL;
   }
      
   for(uplink = MyUplink; uplink; uplink = uplink_next) {
      uplink_next = uplink->next;
      xfree(uplink);
   }
   MyUplink = NULL;
      
   if (MyConsole) {
      if (MyConsole->channel)
         xfree(MyConsole->channel);
      xfree(MyConsole);
      MyConsole = NULL;
   }
      
   if (MyClient) {
      if (MyClient->hostname)
         xfree(MyClient->hostname);
      if (MyClient->realname)
         xfree(MyClient->realname);
      xfree(MyClient);
   }
    
   if (MyPerform) {
      xfree(MyPerform);
      MyPerform = NULL;
   }
      
   for(share = MyShare; share; share = share_next) {
      share_next = share->next;
      for (srv = share->servlist; srv; srv = srv_next) {
         srv_next = srv->next;
         xfree(srv);
      }
      xfree(share);
   }
   MyShare = NULL;
      
   if (MyNoScan) {
      IPrangelist *iprl, *iprl_next;
      for (iprl = MyNoScan->ips; iprl; iprl = iprl_next) {
         iprl_next = iprl->next;
         xfree(iprl);
      }
      for (srv = MyNoScan->residences; srv; srv = srv_next) {
         srv_next = srv->next;
         xfree(srv);
      }
      xfree(MyNoScan);
      MyNoScan = NULL;
   }
   
   if (MyLog) {
      xfree(MyLog);
      MyLog = NULL;
   }
   
   for (tgt = MyTarget; tgt; tgt = tgt_next) {
      tgt_next = tgt->next;
      xfree(tgt);
   }
   MyTarget = NULL;
   
   if (MyGline) {
      Relay_s *r, *r_next;
      for (r = MyGline->relay; r; r = r_next) {
         r_next = r->next;
         xfree(r);
      }
      xfree(MyGline);
      MyGline = NULL;
   }
}


Boolean
conf_reload()
{
   Boolean res;
   
   /* Backup structs (in case of reload failed) */
   Me_s *       MeSave        = Me;
   Uplink_s *   MyUplinkSave  = MyUplink;
   Console_s *  MyConsoleSave = MyConsole;
   MyClient_s * MyClientSave  = MyClient;
   Perform_s *  MyPerformSave = MyPerform;
   Share_s *    MyShareSave   = MyShare;
   NoScan_s *   MyNoScanSave  = MyNoScan;
   Log_s *      MyLogSave     = MyLog;
   Target_s *   MyTargetSave  = MyTarget;
   Gline_s *    MyGlineSave   = MyGline;
   
   res = conf_load();
   
   if (res) {
      
      /* Configuration loading successful, free previous backup'd structs */
      Uplink_s *uplink, *uplink_next;
      Share_s *share, *share_next;
      Target_s *tgt, *tgt_next;
      Serverlist *srv, *srv_next;
      
      if (MeSave)
         xfree(MeSave);
      
      for(uplink = MyUplinkSave; uplink; uplink = uplink_next) {
         uplink_next = uplink->next;
         xfree(uplink);
      }
      
      if (MyConsoleSave) {
         if (MyConsoleSave->channel)
            xfree(MyConsoleSave->channel);
         xfree(MyConsoleSave);
      }
      
      if (MyClientSave) {
         if (MyClientSave->hostname)
            xfree(MyClientSave->hostname);
         if (MyClientSave->realname)
            xfree(MyClientSave->realname);
         xfree(MyClientSave);
      }
      
      if (MyPerformSave)
         xfree(MyPerformSave);
      
      for(share = MyShareSave; share; share = share_next) {
         share_next = share->next;
         for (srv = share->servlist; srv; srv = srv_next) {
            srv_next = srv->next;
            xfree(srv);
         }
         xfree(share);
      }
      
      if (MyNoScanSave) {
         IPrangelist *iprl, *iprl_next;
         for (iprl = MyNoScanSave->ips; iprl; iprl = iprl_next) {
            iprl_next = iprl->next;
            xfree(iprl);
         }
         for (srv = MyNoScanSave->residences; srv; srv = srv_next) {
            srv_next = srv->next;
            xfree(srv);
         }
         xfree(MyNoScanSave);
      }
      
      for (tgt = MyTargetSave; tgt; tgt = tgt_next) {
         tgt_next = tgt->next;
         xfree(tgt);
      }
      
      if (MyGlineSave) {
         Relay_s *r, *r_next;
         for (r = MyGlineSave->relay; r; r = r_next) {
            r_next = r->next;
            xfree(r);
         }
         xfree(MyGlineSave);
      }
      
   } else {
      /* Configuration loading failed, retrive previous settings... */
      Me = MeSave;
      MyUplink = MyUplinkSave;
      MyConsole = MyConsoleSave;
      MyClient = MyClientSave;
      MyPerform = MyPerformSave;
      MyShare = MyShareSave;
      MyNoScan = MyNoScanSave;
      MyLog = MyLogSave;
      MyTarget = MyTargetSave;
      MyGline = MyGlineSave;
   }
   
   return res;
}

/* conf.c */
