/* Pxys - An IRC ProxyScan, m_array.c
 * 
 * Copyright (C) 2001,2002 Stephane Thiell <mbuna@undernet.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * 
 *
 * Thread-safe IRC remote clients array functions.
 *
 * $Id: m_array.c,v 1.49 2002/11/15 19:50:56 mbuna Exp $
 */


#include "config.h"

#include <sys/time.h> // tmp

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#ifdef HAVE_PTHREAD_H
#include <pthread.h>
#endif

#include "m_array.h"
#include "m_clientq.h"
#include "m_structs.h"
#include "m_log.h"
#include "m_cache.h"
#include "m_glineq.h"
#include "m_noscan.h"
#include "m_proxycache.h"
#include "m_share.h"
#include "m_scanner.h"

#include "ms_parser.h"
#include "ms_globals.h"
#include "ms_msgout.h" /* 1.6 */
#include "ms_netutils.h"
#include "ms_network.h" /* 1.6 */

#include "conf.h"
#include "debug.h"
#include "hashtab.h"
#include "recycle.h"
#include "misc.h"
#include "numnicks.h"
#include "strnlwrcpy.h"
#include "typedefs.h"
#include "yxx.h"
/* xmalloc at the bottom of include files */
#include "xmalloc.h"


/* Global variables. */

long clients_counter = 0; /* number of call of array_addclient */

long ipa_current = 0;              /* Current size of the array */
long ipa_max = 0;                  /* Maximum size of the array */
long found_in_cache = 0;

extern long totalcached;

htab *mHC = NULL;
extern htab *mHS;
extern server_handle_t mSA;
extern scanstats_s *scanstats;
extern time_t startscan_time;

/* Using mutex to keep the array synced without conflict with the threads.
 * mutex_link is the main mutex which protects the array critical section (when
 * the array linking is modified). Threads which want to check the
 * array need to wait until these zone are unlocked.
 */
pthread_mutex_t mutex_link;
pthread_mutex_t mutex_shutdown;
pthread_mutex_t mutex_scanext;

extern scanthread_s *scanthread;


static int array_fetchitem_clientq(struct Client **cliH);



/* init the array: mutexes and clients hash table.
 */
void
array_init()
{
   /* Array mutex lockers */
   pthread_mutex_init(&mutex_link, NULL);
   pthread_mutex_init(&mutex_shutdown, NULL);
   pthread_mutex_init(&mutex_scanext, NULL);
   
   /* Create main clients hash table */
   mHC = jenkins_hcreate(2);
}

/* destructor.. hrm.. yeah it's still written in C,
 * but would be nice in C++.
 */
void
array_destroy()
{
   if (mHC) {
      jenkins_hdestroy(mHC);
      mHC = NULL;
      pthread_mutex_destroy(&mutex_link);
      pthread_mutex_destroy(&mutex_shutdown);
      pthread_mutex_destroy(&mutex_scanext);
   }
}

/* addclient is always called by the same (main) thread
 * (the one the IRC socket is).
 */
void
array_addclient(
  const char *nick,       /* Client nickname */
  const char *mode,       /* Client mode; can be NULL */
  const char *b64IP,      /* Client base64 IP */
  const char *numnick,    /* Client numeric */
  int ident)              /* 1 if idented, 0 if not (~user) */
{
   yxx target_yxx;
   int cached = 0;
   proxycacheItem *citem;
   struct Server *sptr = NULL;
   struct Client *cptr = (struct Client *) xmalloc(sizeof(struct Client));
   
   /* Fill Client struct... */
   target_yxx = yxx_to_int(numnick);
   cptr->nserv = target_yxx.server_n;
   cptr->nnick = target_yxx.client_n;
   strnlwrcpy(cptr->nick, nick, NICKLEN);
   cptr->nick[NICKLEN] = '\0';
   
   cptr->flags = FLAG_NULL;
   if (ident)
      cptr->flags |= FLAG_IDENT;
   if (mode && strchr(mode, 'o'))
      cptr->flags |= FLAG_OPER;
   
   cptr->ip.s_addr = htonl(base64toint(b64IP));
   cptr->last = 0;
   
   pthread_mutex_lock(&mutex_link);
   
   /* Find server; get clients array; link. */
   if (target_yxx.server_n < NN_MAX_SERVER)
      sptr = mSA[target_yxx.server_n];
   if (NULL == sptr) {
      log_error("array_addclient: server not found (%u)\n", target_yxx.server_n);
      pthread_mutex_unlock(&mutex_link);
      return;
   }
   
   sptr->cliH[target_yxx.client_n & sptr->nn_mask] = (void*)cptr;
   ++(sptr->clients);
   /* If needed, update the maxclients record for this server */
   if (sptr->maxclients < sptr->clients)
      sptr->maxclients = sptr->clients;
   
   hadd(mHC, cptr->nick, strlen(cptr->nick) + 1, (void*)cptr);
   
   ++ipa_current;
   if (ipa_max < ipa_current)
      ipa_max = ipa_current;   /* Only here for fun */
   
   /* Another thingy... proxycache, new in 1.5
    * Rebuilt in 1.6 to support glineq...
    */
   
   if ((cached = proxycache_check(&cptr->ip, &citem))) {
      if (is_bursting(sptr)) {
         char *typeproxy, *reason;
         get_proxy_type(citem->type, citem->port, 1, &typeproxy, &reason);
         if (!reason)
            log_error("array_addclient: get_proxy_type() failed (%x)", citem->type);
         if (reason && glineq_push(&citem->addr, citem->type)) {
#ifdef CONSOLE_SHOWCACHED
            if (citem->type == PROXY_TYPE_HTTPPROXY)
               client_sendmsgto_console(
	              "%s *@%s [1+] %s %u (cached+queued)", PROXY_GLINE_CMD,
                  inet_ntoa(citem->addr), typeproxy, citem->port);
            else
               client_sendmsgto_console(
	              "%s *@%s [1+] %s (cached+queued)", PROXY_GLINE_CMD,
                  inet_ntoa(citem->addr), typeproxy);
#endif
         }
      } else {
         /* Warning: avoid dead lock here! */
         pthread_mutex_unlock(&mutex_link);      /* UNLOCK */
         proxy_toast(&cptr->ip, citem->type, 1);
         pthread_mutex_lock(&mutex_link);        /* LOCK */
      }
      totalcached++;
   }
   
   ++clients_counter;
   
   /* ClientQ support: recent connected users */
   if (!cached)
      clientq_queue(&target_yxx);
   
   pthread_mutex_unlock(&mutex_link);
}


/* delclient is always called by the same (main) thread.
 * Used to delete only 1 client (after a QUIT or KILL message).
 */
void
array_delclient(const char *numnick)
{
   struct Server *sptr;
   struct Client *cptr;
   yxx target_yxx;
   
   target_yxx = yxx_to_int(numnick);
   
   pthread_mutex_lock(&mutex_link);
   
   sptr = mSA[target_yxx.server_n];
   if (NULL == sptr) {
      pthread_mutex_unlock(&mutex_link);
      log_error("array_delclient: server not found (%d)", target_yxx.server_n);
      return;
   }
   
   cptr = sptr->cliH[target_yxx.client_n & sptr->nn_mask];
   if (NULL == cptr) {
      pthread_mutex_unlock(&mutex_link);
      log_error("array_delclient: client not found (%d %d)",
           target_yxx.server_n, target_yxx.client_n);
      return;
   }
   
   if (hfind(mHC, cptr->nick, strlen(cptr->nick) + 1)) {
      /* Alright, user found in hash table too. */
      if (hdel(mHC) == FALSE)
         log_error("array_delclient: hdel failed");
   } else {
      /* This should really not happen... */
      log_error("array_delclient: hashed client not found (%d %d) nick: %s (num: %s)",
           target_yxx.server_n, (target_yxx.client_n & sptr->nn_mask),
	   cptr->nick, numnick);
   }

   xfree(cptr);
   
   sptr->cliH[target_yxx.client_n & sptr->nn_mask] = NULL;
   sptr->clients--;
   --ipa_current;
   
   pthread_mutex_unlock(&mutex_link);
}


/* array_purgeserv is always called by the same (main) thread.
 * Used to free all users from a specified server numeric.
 *
 * I rewrote a low level dealing of the hash table; that's why:
 *
 * "The most unusual thing about this module (hashtab.c) is that it maintains
 *  a current position. This means you can't have a dangling pointer into
 *  it. If you position on something, then delete it, the position
 *  automatically moves on to some other item. On the other hand, it
 *  also means it's hard to do nested loops over all the items in the
 *  table, because there is only one position at a time. " -- Bob Jenkins
 * 
 *
 * --mbuna 20010327
 *
 * mutex_link protected from ms_network:do_squit()
 */

void
array_purgeserv(unsigned char yy)
{
   struct Server *sptr;
   ub4 i, end, counter;
   
   Debug((DBLMAX, "array_purgeserv> yy == %u", yy));
   
   sptr = mSA[yy];
   
   end = (ub4)1<<(mHC->logsize);
   counter = 0;
   for (i = 0; i < end;  ++i) {
      hitem *h, *hnext;
      for (h=mHC->table[i]; h; h=hnext) {
         struct Client *cptr;
         
         hnext = h->next;
         
         cptr = (struct Client *)h->stuff;
         
         if (cptr->nserv == yy) {
            
            hitem **ip;   /* a counter */
            /* hash table: remove item from its list */
            for (ip = &mHC->table[i]; *ip != h; ip = &(*ip)->next)
               ;
            *ip = (*ip)->next;
            --(mHC->count);
            
             /* hash table: recycle the deleted hitem node */
            redel(mHC->space, h);
            
            sptr->cliH[cptr->nnick & sptr->nn_mask] = NULL;
            sptr->clients--;
            --ipa_current;
            
            xfree(cptr);
         }
      }
   }
   
   /* hsanity(mHC); */
   
   /* warning: ipos not in apos -> adjust position */
   (void) hnbucket(mHC);
}


/* Called from the main thread only. */
void
array_nickchange(const char *numnick, const char *newnick)
{
   struct Server * sptr;
   struct Client * cptr;
   yxx target_yxx;
   
   target_yxx = yxx_to_int(numnick);
   
   sptr = mSA[target_yxx.server_n];
   if (NULL == sptr) {
      log_error("array_nickchange: Server #%d not found!", target_yxx.server_n);
      return;
   }
   
   /* Server found. */

   cptr = sptr->cliH[target_yxx.client_n & sptr->nn_mask];
   if (NULL == cptr) {
      log_error("array_nickchange: Client ID %s (%d) not found!",
           numnick, target_yxx.client_n & sptr->nn_mask);
      return;
   }
   
   /* Same nick (like warlordz -> WaRlOrDz) ? */
   if (stricmp(cptr->nick, newnick) == 0)
      return;
   
   /* Client found. */
   pthread_mutex_lock(&mutex_link);
   
   if (hfind(mHC, cptr->nick, strlen(cptr->nick) + 1)) {
      /* Remove it from hash table, as we hash on nicks. */
      hdel(mHC);
      /* Update Client struct. Only deal with lowercase nicks. */
      strnlwrcpy(cptr->nick, newnick, NICKLEN);
      cptr->nick[NICKLEN] = '\0';
      
      /* Add it to hash table. */
      hadd(mHC, cptr->nick, strlen(cptr->nick) + 1, (void*)cptr);
   } else
      log_error("array_nickchange: hfind failed, nick %s not found on %s",
           cptr->nick, sptr->name);
   
   pthread_mutex_unlock(&mutex_link);
}


/* array_fetchitem is called by EVERY scan threads!
 * Used to get/set the next targets for the thread th.
 * Pre-condition: th != NULL.
 */
void
array_fetchitem(scanthread_s * th)
{
   int i;
   
   th->counter = 0;
   
   /* Cleaning */
   if (scanth_isrunning(th))
      for (i = 0; i < SCANS_PER_THREAD; i++)
         th->item[i].state = SIT_INACTIVE;
            
   pthread_mutex_lock(&mutex_link); /* suspend any clients list
                                     * or clientq changes during the fetch
                                     */
   
   for (i = 0; i < SCANS_PER_THREAD; i++) {
      struct Client *cli;
      int result = array_fetchitem_clientq(&cli);
         
      if (result) {
         th->item[i].nserv = cli->nserv;
         th->item[i].nnick = cli->nnick;
         memcpy(&th->item[i].addr, &cli->ip, sizeof(struct in_addr));
         th->item[i].state = SIT_ACTIVE;
         th->counter++;
      } else
         /* no more clients in clientq, this is the end */
         break;
   }
   if (th->counter == 0) {
      /* We didn't found any clients in clientq, and we are sync? It simply
       * means no new client has connected. Thus, this thread is going to wait now.
       */
      if (scanth_isrunning(th)) {
         /* Change scanthread state */
         scanstats->running_threads--;
         scanstats->waiting_threads++;
         scanth_setwaiting(th);
      }
      /* else.. This thread was already "waiting" for something to scan */
   } else {
      /* We found at least 1 IP to scan */
      if (scanth_iswaiting(th)) {
         scanstats->running_threads++;
         scanstats->waiting_threads--;
      }
      scanth_setrunning(th);
   }
   pthread_mutex_unlock(&mutex_link);
}

/* RETURNS
 * 1 if a valid client to scan is found
 * 0 if there is no more client to scan in clientq
 */
static int
array_fetchitem_clientq(struct Client **cliH)
{
   int noscan;
   time_t now, cached;
   
   time(&now);
   
   for (;;) {
      struct Client *cli;
      struct Server *srv;
      yxx* cliyxxp;
      
      /* clientQ dequeue... we break if no more clients are in queue. */
      if (!clientq_deque(&cliyxxp))
         break;
      
      srv = mSA[cliyxxp->server_n];
      
      /* srv can be NULL, eg. if it splits between client's
       * enqueue and dequeue.
       */
      if (!srv)
         continue;
      
      cli = srv->cliH[cliyxxp->client_n & srv->nn_mask];
      /* cli can be NULL, eg. if the client quits before the dequeue. */
      if (!cli)
         continue;

      /* Verify our configuration scan exceptions.
       */
      check_noscan_server(cli->nserv, &noscan);
      
      /* And if we're working with another proxyscanner...
       */
      if (!noscan)
         check_shared_server(cli->nserv, &noscan);

      if (noscan)
         continue;

      if (cli->last == 0) {
         /* client not checked */

         cli->last = now;

         /* "protected" by noscan ? */
         check_noscan_ip(&cli->ip, &noscan);

         if (noscan)
            continue;

         /* Check cache */
         cached = cache_fetch(&cli->ip);

         if (!cached || (now - cached > MyPerform->recheck)) {
            cache_insert(&cli->ip, now);
            *cliH = cli;
            return 1;
         } else
         	found_in_cache++;
      }
   }
   /* *boom* nothing valid found */
   return 0;
}


/* Might be called by the other threads
 * BTW increases the currtoasted count on the related servers
 */
long
array_countclient_from_ip(struct in_addr *addr)
{
   long counter = 0;
   struct Client *cptr;

   pthread_mutex_lock(&mutex_link);

   if (hfirst(mHC)) do {
       cptr = (struct Client *)hstuff(mHC);
       if (addr->s_addr == cptr->ip.s_addr) {
          ++counter;
          ((struct Server *)mSA[cptr->nserv])->currtoasted++;
       }
    } while(hnext(mHC));
    
    pthread_mutex_unlock(&mutex_link);
    
    return counter;
}

/* Have to be called from the 'main thread', the same thread which is calling
 * addclient and delclient (the IRC socket thread). */
struct Client *
array_findclient_from_numeric(const char *numnick)
{
   struct Server *sptr;
   yxx client_yxx;
   
   client_yxx = yxx_to_int(numnick);
   sptr = mSA[client_yxx.server_n];
   if (NULL == sptr)
      return NULL;
   
   return sptr->cliH[client_yxx.client_n & sptr->nn_mask];
}

/* Called from the main thread only. */
struct Client *
array_findclient_from_nick(const char *nick)
{
   char tmp_nick[NICKLEN+1];
   
   strnlwrcpy(tmp_nick, nick, NICKLEN);
   tmp_nick[NICKLEN] = '\0';
   
   if (hfind(mHC, tmp_nick, strlen(nick) + 1))
      return (struct Client *)hstuff(mHC);
   
   return NULL;
}




/* m_array.c */
